# Copyright (C) 2007 - 2010 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
# Bjoern Ricks    <bjoern.ricks@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Collection of subversion utility code"""

import os
import shutil
import re
import StringIO

from lxml import etree

import run
from cmdexpand import cmdexpand
from util import extract_value_for_key


class SubversionError(Exception):

    """Base class for subversion specific errors raised by TreePKG"""


class SubversionUrlMismatchError(SubversionError):

    """The repository URL does not match the URL of a working copy"""

class Svn:

    def __init(self):
        pass
    
    def get_tag_date(self, tag_url):
        match = re.search(r"/enterprise[^.]*\.[^.]*\."
                          r"(?P<date>[0-9]{8})\.(?P<baserev>[0-9]+)/",
                          tag_url)
        if match:
            date = match.group("date")
            baserev = match.group("baserev")
            return date
        else:
            raise RuntimeError("Cannot determine tag date from %r"
                               % tag_url)


    def list_url(self, url):
        """Runs svn list with the given url and returns files listed as a list"""
        output = run.capture_output(cmdexpand("svn list $url", **locals()))
        return output.splitlines()

    def checkout(self, url, localdir, revision=None, recurse=True):
        """Runs svn to checkout the repository at url into the localdir"""
        args = []
        if revision:
            args.extend(["--revision", revision])
        if not recurse:
            args.append("-N")
        run.call(cmdexpand("svn checkout -q @args $url $localdir", **locals()))

    def update(self, localdir, revision=None, recurse=True):
        """Runs svn update on the localdir.
        The parameter revision, if given, is passed to svn as the value of
        the --revision option.
        """
        args = []
        if revision:
            args.extend(["--revision", revision])
        if not recurse:
            args.append("-N")
        run.call(cmdexpand("svn update -q @args $localdir", **locals()))

    def export(self, src, dest, revision=None, recurse=True):
        """Runs svn export src dest"""
        args = []
        if revision:
            args.extend(["--revision", revision])
        if not recurse:
            args.append("-N")
        run.call(cmdexpand("svn export -q @args $src $dest", **locals()))

    def last_changed_revision(self, svn_working_copy):
        """return the last changed revision of an SVN working copy as an int"""
        # Make sure we run svn under the C locale to avoid localized
        # messages
        env = os.environ.copy()
        env["LANG"] = "C"

        output = run.capture_output(cmdexpand("svn info $svn_working_copy",
                                              **locals()),
                                    env=env)
        str_rev = extract_value_for_key(output.splitlines(), "Last Changed Rev:")
        if str_rev is None:
            raise SubversionError("Cannot determine last changed revision for %r"
                                  % svn_working_copy)
        return str_rev

    def url(self, url_or_working_copy):
        """Returns the URL used for the working copy in svn_working_copy"""
        # Make sure we run svn under the C locale to avoid localized
        # messages
        env = os.environ.copy()
        env["LANG"] = "C"

        output = run.capture_output(cmdexpand("svn info $url_or_working_copy",
                                              **locals()),
                                    env=env)
        return extract_value_for_key(output.splitlines(), "URL:")

    def log_xml(self, url, base_revision):
        """Return the log in XML of the repository at url from base_revision to HEAD
        """
        args = ["--revision", str(base_revision) + ":HEAD",
                "--verbose",
                "--xml"]
        return run.capture_output(cmdexpand("svn log @args $url", **locals()))


    def extract_tag_revisions(self, xml_log):
        """Extracts the revisions which changed an SVN tag since its creation
        This includes the revision which created the tag and all subsequent
        changes.  The xml_log parameter should contain the xml-Version of
        the SVN log of the tag that includes at least the revision that
        created the tag and all the newer revisions.
        """
        tree = etree.parse(StringIO.StringIO(xml_log))
        tag_revisions = tree.xpath("logentry/@revision"
                                   "[.>=../../logentry/@revision"
                                   "[../paths/path[@copyfrom-path]]]")
        return tag_revisions


class SvnTagDetector(object):

    """Class to automatically find SVN tags and help package them

    The tags are found using three parameters:
      url -- The base url of the SVN tags directory to use
      pattern -- A regular expression matching the subdirectories to
                 consider in the tag directory specified by the url
      subdir -- A subdirectory of the directory matched by pattern to
                export and use to determine revision number

    The subdir parameter is there to cope with the kdepim enterprise
    tags.  The URL for a tag is of the form
    .../tags/kdepim/enterprise4.0.<date>.<rev> .  Each such tag has
    subdirectories for kdepim, kdelibs, etc.  The url and pattern are
    used to match the URL for the tag, and the subdir is used to select
    which part of the tag is meant.

    The subdir also determines which SVN directory's revision number is
    used.
    """

    def __init__(self, url, pattern, subdir):
        self.url = url.rstrip("/")
        self.pattern = re.compile(pattern)
        self.subdir = subdir
        self.tagurl = None
        self.svn = Svn()

    def list_tags(self):
        matches = []
        if self.url:
            for tag in self.svn.list_url(self.url):
                if self.pattern.match(tag.rstrip("/")):
                    matches.append(tag)
        return sorted(matches)

    def newest_tag_revision(self):
        """Determines the newest tag revision and returns (tagurl, revno,
        tagsubdira). If no tag can be found, the method returns the 
        tuple (None, None, None).
        """
        candidates = self.list_tags()
        urlrev = (None, None, None)
        if candidates:
            newest = candidates[-1]
            urlrev = self.determine_revision(self.url + "/" + newest,
                                             self.subdir)
        self.tagurl = urlrev[0]
        return urlrev

    def determine_revision(self, baseurl, subdir):
        urlrev = (None, None, None)
        baseurl = baseurl.rstrip("/")
        revision_url = baseurl + "/" + subdir
        try:
            revision = self.svn.last_changed_revision(revision_url)
            urlrev = (baseurl + "/" + subdir, revision, subdir)
        except SubversionError:
            pass
        return urlrev
    
    def get_tag_date(self):
        return self.svn.get_tag_date(self.tagurl)

class GitError(Exception):

    """Base class for Git specific errors raised by TreePKG"""

class Git:

    def __init__(self):
        pass

    def checkout(self, url, localdir, branch=None):
        """Clones the repository at url into the localdir"""
        run.call(cmdexpand("git clone -q $url $localdir", **locals()))
        if branch:
            run.call(cmdexpand("git checkout -q --track -b local $branch",
                                **locals()), cwd=localdir)

    def update(self, localdir, revision=None):
        """Runs git pull on the localdir."""
        run.call(cmdexpand("git pull -q"), cwd=localdir)

    def export(self, src, dest):
        """Exports the local branch from src to dest"""
        dest = dest + os.sep
        run.call(cmdexpand("git checkout-index -a -f --prefix=$dest", **locals()),
                 cwd=src)

    def last_changed_revision(self, git_working_copy):
        """Return the SHA1 sum of the latest commit"""
        output = run.capture_output(cmdexpand("git rev-parse HEAD"),
                                    cwd=git_working_copy)
        if output is None:
            raise GitError("Cannot determine last changed revision for %r"
                           % git_working_copy)
        return output.strip()

    def get_tag_date(self, tagurl):
        """ TODO """
        return None

class GitTagDetector:

    """Class to detect tags from a git repository

    The tags are found using the parameters:
      url -- The url of the git repository to use
      pattern -- A regular expression matching the tags
    """

    def __init__(self, url, pattern, localdir):
        self.url = url
        self.pattern = pattern
        self.localdir = localdir
        self.git = Git()
        if not os.path.exists(localdir):
            self.git.checkout(url, localdir)
        self.git.update(localdir)

    def list_tags(self):
        output = run.capture_output(cmdexpand("git tag -l $pattern",
            pattern=self.pattern), cwd=self.localdir)
        tags = output.splitlines()
        return sorted(tags)

    def get_revision(self, refname="HEAD"):
        """Return the SHA1 sum of the latest commit"""
        output = run.capture_output(cmdexpand("git rev-parse $refname",
            refname=refname), cwd=self.localdir)
        if output is None:
            raise GitError("Cannot determine revision for %r"
                           % self.localdir)
        return output.strip()

    def newest_tag_revision(self):
        candidates = self.list_tags()
        urlrev = (None, None, None)
        if candidates:
            newest = candidates[-1]
            try:
                rev = self.get_revision(newest)
                urlrev = (newest, rev, self.localdir)
            except GitError:
                pass
        return urlrev

    def tag_pkg_parameters(self, tag_name):
        # FIXME: Don't hardcore regex
        #match = re.search(r"enterprise[^.]*\.[^.]*\."
        match = re.search(r"enterprise[^.]*\.[^.]*"
                          r"(?P<date>[0-9]{8})",
                          tag_name)
        if match:
            date = match.group("date")
            return (date, 1)
        else:
            raise GitError("Cannot determine tag parameters from %s"
                               % tag_name)
    def get_tag_date(self):
         return self.git.get_tag_date(self.url)
